/* PTMovies	-	Display Panorama Movies
   Copyright (C) 2000 - Helmut Dersch  der@fh-furtwangen.de
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.net.*;
import java.net.URLConnection;
import java.util.*;
import java.text.*;
import java.io.*;
import java.lang.reflect.*;
import java.lang.*;


public class ptmviewer extends Applet
{
	String PTViewer = "ptviewer";
	ptviewer pv = null;
	String filename = "image";
	String order=null;
	
	int y_LU[] = null;
	int w0[]   = null;
	int w1[]   = null;
	
	boolean hasCube 	= false;
	boolean hasPano 	= false;
	boolean hasHotspots = false;
	boolean hasJpeg 	= false;
	boolean hasTiff 	= false;


	public ptmviewer(){}


	public ptmviewer( ptviewer p, String properties ){
		pv = p;
		setStub(new ptstub(pv, properties ));
	}	

	
   	public void init(){
   		String s;
   		s = getParameter( "file"); 	if( s!=null ) 	filename = s;
  		s = getParameter( "order"); 	if( s!=null ) 	order = s;
	}
	
	public void start(){
 		if( pv!=null){
 			loadQTVR();
		}
   	}
   	
	
	void loadQTVR(){
		byte[] qtvr = null;
		// Read file into memory

		pv.percent[0] = 0;
		pv.repaint();
		qtvr = pv.file_read( filename, pv.percent );
			
		if(qtvr == null){
			pv.fatal = true;
			return;
		}
		
		// Check if this is a QTVR-Movie and if yes, what type
		
		if( !check_qtvr( qtvr ) ){
			System.out.println("This doesn't seem to be a Quicktime VR-file");
			pv.fatal = true;
			return;
		}
		
		if( !hasJpeg ){
			System.out.println("Images in QTVR-file must use Photo-JPEG format.");
			pv.fatal = true;
			return;
		}
				
		
		// Read user data and set properties
		
		set_qtvr_properties( get_qtvr_properties( qtvr ) );
		
		
		if( hasPano ){
			if( hasCube ){
				pv.show_pdata = false;
				displayCubicQTVR( qtvr );
			}else{
				displayCylindricalQTVR( qtvr );
			}
		}else{					
			pv.show_pdata = false;
			displayObjectQTVR( qtvr );
		}
	}
				
	
	void displayObjectQTVR( byte[] qtvr ){
		if(qtvr==null) return;
 		
 		// Read images
				
		Image[] frames = get_jpeg_images( qtvr );
		if( frames == null  || frames.length == 0 ){
			System.out.println("Could not read QTVR-file.");
			pv.fatal = true;
			return;
		}
		String s = "{code=ptobject.class}{nhor=" + frames.length + "}";
				
		ptobject a = new ptobject( pv, frames, s);
		pv.app_properties.addElement(s);
		pv.applets.put(s, a);
		a.init();
		a.start();
	}
	
	
	void displayCylindricalQTVR(  byte[] qtvr ){
		if(qtvr==null) return;
 		
 		// Read images
				
		Image[] cyl_tiles = get_jpeg_images( qtvr );
		if( cyl_tiles == null  || cyl_tiles.length == 0 ){
			System.out.println("Could not read QTVR-file.");
			pv.fatal = true;
			return;
		}
		
		// allocate memory
		int pwidth 	= cyl_tiles[0].getHeight(null) * cyl_tiles.length;
		int cheight = cyl_tiles[0].getWidth(null);
		int pheight = (int)((double)pwidth/Math.PI * Math.atan((double)cheight/(double)pwidth * Math.PI) + 0.5);
		
		int[][] pdata = null;
		
		try{
			pdata = new int[pheight][pwidth];
		}catch(Exception e){}
		if(pdata==null){
			System.out.println("Not enough memory to load panorama");
			pv.fatal=true;
			return;
		}

		int[] idata = new int[cyl_tiles[0].getHeight(null) * cheight];
		byte [][]hs = new byte[cyl_tiles.length][];
		for(int i=0; i<cyl_tiles.length; i++) hs[i] = null;
			 	
		if( hasHotspots && hasTiff )
        	hs = get_tiff_hsimages( qtvr, cyl_tiles.length );

		
		for(int i=0; i<cyl_tiles.length; i++){
			pv.percent[0] = 80 + (i+1) * 20 / cyl_tiles.length;
			pv.repaint();
			Image im = cyl_tiles[i];
       		PixelGrabber pg = new PixelGrabber( im, 0, 0, im.getWidth(null), im.getHeight(null), idata, 0, im.getWidth(null) );
        	try { pg.grabPixels(); } 
        	catch (InterruptedException e){
            	return;
        	}
        	insert_cyltile( i, idata, cheight, pdata, hs[i] );
        }
        // clean up
        y_LU = null; w0 = null; w1 = null;
        
        pv.pdata = pdata;
	}

	void insert_cyltile( int i, int[] idata, int cheight, int[][] pdata, byte[] hs ){
		if( idata==null || pdata == null ) return;
		
		int pheight = pdata.length;
		int pwidth  = pdata[0].length;
		int twidth  = idata.length/cheight;
		if( y_LU == null ){	// Set up lookup table
			y_LU = new int[pheight];
			w0	 = new int[pheight];
			w1	 = new int[pheight];
			double d = (double)pwidth /(2.0*Math.PI);
			double di = d * 256.0;
			double h2 = (double)pheight / 2.0 + 0.5;
			int ch2 = cheight * 128;
			for(int y=0; y<pheight; y++){
				int idx = (int)(di * Math.tan(((double)y-h2)/d)) + ch2;
				w1[y] = idx & 0xff;
				w0[y] = 255 - w1[y];
				y_LU[y] = idx >> 8;
				if(y_LU[y] < 0) y_LU[y] = 0;
				if(y_LU[y] >= cheight-1) y_LU[y] = cheight-2;
			}
		}
		
		int it = pwidth - 1 - i*twidth;
		for(int y=0; y<pheight; y++){
			for(int x=0, xc=0; x<twidth; x++,xc+=cheight){
				int xi = xc+y_LU[y];
				int pix0 = idata[xi];
				int pix1 = idata[xi+1];
				int alpha = (hs==null ? 0 : ((int)hs[xi]) << 24) ;
				int r = (((pix0 >> 16) & 0xff) * w0[y] + ((pix1 >> 16) & 0xff) * w1[y]) >> 8;
				if( r > 0xff ) r = 0xff; if(r<0) r=0;
				int g = (((pix0 >>  8) & 0xff) * w0[y] + ((pix1 >>  8) & 0xff) * w1[y]) >> 8;
				if( g > 0xff ) g = 0xff; if(g<0)  g=0;
				int b = (((pix0      ) & 0xff) * w0[y] + ((pix1      ) & 0xff) * w1[y]) >> 8;
				if( b > 0xff ) b = 0xff; if(b<0) b=0;
					 
				pdata[y][it - x] = alpha + (r<<16) + (g<<8) + b;
			}
		}
	}
  
  	void displayCubicQTVR( byte[] qtvr ){
  		if(qtvr==null) return;
 		
 		// Read images
				
		Image[] cubic_faces = get_jpeg_images( qtvr );
		if( cubic_faces == null  || cubic_faces.length != 6 ){
			System.out.println("Could not read QTVR-file.");
			pv.fatal = true;
			return;
		}
				
		String s[] = {"{code=ptzoom.class}{fov=90.2}{pan=180}",
			 		  "{code=ptzoom.class}{fov=90.2}{pan=-90}",
			 		  "{code=ptzoom.class}{fov=90.2}{pan=0}",
					  "{code=ptzoom.class}{fov=90.2}{pan=90}",
			 		  "{code=ptzoom.class}{fov=90.2}{pan=180}{tilt=90}",
			 		  "{code=ptzoom.class}{fov=90.2}{pan=180}{tilt=-90}"};
			 		  		
		int i;
		byte [][]hs = new byte[6][];
		for(i=0; i<6; i++) hs[i] = null;
			 	
		if( hasHotspots && hasTiff )
        	hs = get_tiff_hsimages( qtvr, 6 );

		for(i=0; i<6; i++){
			pv.percent[0] = 80+i*4;
			pv.repaint();

			ptzoom a = new ptzoom( pv, cubic_faces[i], hs[i], s[i]);
			hs[i] = null;
			pv.app_properties.addElement(s[i]);
			pv.applets.put(s[i], a);
			a.init();
			a.start();
			/*while( !a.ready ){		
				try{ 
					Thread.sleep( 200 );
				}catch (InterruptedException ie) {}
			}*/
		}
	}
 		
   	Properties get_qtvr_properties( byte[] b ){
   		Properties p = new Properties();
   		
   		byte ID[] = { (byte)0x70, (byte)0x6c, (byte)0x75, (byte)0x67 }; // 'plug'
   		
   		int count = b.length- ID.length;
   		
   		for(int i=0; i<count; i++){
   			if( b[i] == ID[0] && b[i+1] == ID[1] && b[i+2] == ID[2] && b[i+3] == ID[3] ){ // We found a value=property pair
   				i+=4;
   				int k=i;
   				while( k < b.length && b[k] != (byte)0 ) k++;
   				if(k>i){
   					byte[] cs = new byte[k-i];
   					System.arraycopy(b,i,cs,0,k-i);
   					String s = new String( cs );
   					int e = s.indexOf('=');
   					if( e > 0 && e < s.length() - 1){
   						p.put( s.substring(0,e).toLowerCase(), s.substring(e+1) );
   						//System.out.println("Found property " + s.substring(0,e).toLowerCase() + " with value " + s.substring(e+1) );
   					}
   					cs = null; s = null;
   				}
   			}
   		}
   		return p;
   	}
	
	// extract jpeg images from array b
	// Choose the largest images with equal size;
	Image[] get_jpeg_images( byte[] b ){
		byte ID[] = { (byte)0xff, (byte)0xd8, (byte)0xff, (byte)0xe0, 
					  (byte)0x00, (byte)0x10, (byte)0x4a, (byte)0x46, 
					  (byte)0x49, (byte)0x46, (byte)0x00 };

		// Determine number of images
		int i,k;
		
		Vector startIndex = new Vector();

		int count = b.length - ID.length;
		
		for(i=0; i<count; i++){
			if( b[i] == ID[0] && b[i+1] == ID[1] && b[i+2] == ID[2] && b[i+3] == ID[3] 
			 && b[i+4] == ID[4] && b[i+5] == ID[5] && b[i+6] == ID[6] && b[i+7] == ID[7]
		 	 && b[i+8] == ID[8] && b[i+9] == ID[9] && b[i+10] == ID[10]){ // Start of JPEG
		 	 		startIndex.addElement( new Integer(i));
			}
		}

		if(startIndex.size() == 0 ){
			return null;
		}

		Vector stopIndex  = new Vector();
		
		for(k=0; k<startIndex.size(); k++){
			int first = ((Integer)startIndex.elementAt(k)).intValue() + ID.length;
			int last;
			if( k < startIndex.size() - 1 ){
				last  = ((Integer) startIndex.elementAt(k+1)).intValue() - 2;
			}else
				last = b.length - 2;
				
			for(i=first; i<=last; i++){
				if( b[i] == (byte)0xff && b[i+1] == (byte)0xd9 ){ // End of JPEG
		 	 		stopIndex.addElement( new Integer(i+2)  );
		 	 		break;
		 	 	}
			}
			if( stopIndex.size() != k+1 ){
				System.out.println("Could not read QTVR file");
				return null;
			}
		}
		
		// System.out.println("Found " + startIndex.size() + " Panos");
		
				
		Vector vim = new Vector();			
			
		for(i=0; i<startIndex.size(); i++){
			int start = ((Integer)startIndex.elementAt(i)).intValue();
			int stop  = ((Integer) stopIndex.elementAt(i)).intValue();
			count = stop - start;
			byte[] buf = new byte[count];
			System.arraycopy(b,start,buf,0,count);
			vim.addElement( pv.bufferToImage( buf )  );
		}

		// Remove all small images
		
		int maxSize = 0, size_i;
		for(i=0; i<vim.size(); i++){
			Image p = (Image)vim.elementAt(i);
			size_i = p.getWidth(null) * p.getHeight(null);
			if( size_i > maxSize ) maxSize = size_i;
		}
		
		for(i=0; i<vim.size(); i++){
			Image p = (Image)vim.elementAt(i);
			size_i = p.getWidth(null) * p.getHeight(null);
			if( size_i < maxSize ){
				vim.removeElementAt(i--);
			}
		}		
		
		// Reorder images
		if(order != null)
			ordertiles(vim, order);
		
		Image[] im= new Image[vim.size()];
		vim.copyInto( im );
				
		return im;

	}
	
	void ordertiles(Vector v, String s){
		Vector vc = (Vector) v.clone();
		if(pv==null) return;
		for(int i=0; i<vc.size(); i++){
			String arg = pv.getArg( i, s );
			if( arg != null ){
				v.setElementAt( vc.elementAt( Integer.parseInt(arg) ), i );
			}
		}
	}

	byte[][] get_tiff_hsimages( byte[] b, int numIm ){
		byte[][] hs = new byte[numIm][];
		int fp = 0;
		int fi = 0;
		
		for(int i=0; i<numIm; i++)
			hs[i] = null;
		
		for(int i=0; i<b.length; i++){
			int offset = readtiffHeader(b, i);
			if( offset > 0 ){
				if( readtiffdirectory( b, i + offset) ){
					hs[fi] = readtiffdata( b );
					if(hs[fi] != null)
						fi++;
					if(fi >= numIm) 
						break;
				}
				i += offset;
			}
		}
		// System.out.println("Found " + fi + " hotspot faces");
		
		return hs;
	}
					
					
	
	////////////// TIFF-stuff: read and decode 256color compressed tiffs

	//constants
	static final int UNSIGNED = 1;
	static final int SIGNED = 2;
	static final int FLOATING_POINT = 3;

	//field types
	static final int SHORT = 3;
	static final int LONG = 4;


	// tags
	public static final int IMAGE_WIDTH = 256;
	public static final int IMAGE_LENGTH = 257;
	public static final int BITS_PER_SAMPLE = 258;
	public static final int COMPRESSION = 259;
	public static final int PHOTO_INTERP = 262;
	public static final int STRIP_OFFSETS = 273;
	public static final int ROWS_PER_STRIP = 278;
	public static final int STRIP_BYTE_COUNT = 279;
	public static final int X_RESOLUTION = 282;
	public static final int Y_RESOLUTION = 283;
	public static final int RESOLUTION_UNIT = 296;
	public static final int COLOR_MAP = 320;
	public static final int SAMPLES_PER_PIXEL = 277;

	
	private boolean littleEndian;
	private int tiff_fp, tiff_start;
	private int tiff_width, tiff_height, tiff_rps;
	private int[] tiff_so;
	private int[] tiff_sbc;
	

	final int getInt(byte[] b) throws IndexOutOfBoundsException{
		int b1 = b[tiff_fp++] & 0xff;
		int b2 = b[tiff_fp++] & 0xff;
		int b3 = b[tiff_fp++] & 0xff;
		int b4 = b[tiff_fp++] & 0xff;
		if (littleEndian)
			return ((b4 << 24) + (b3 << 16) + (b2 << 8) + (b1 << 0));
		else
			return ((b1 << 24) + (b2 << 16) + (b3 << 8) + b4);
	}


	final int getShort(byte[] b) throws IndexOutOfBoundsException{
		int b1 = b[tiff_fp++] & 0xff;
		int b2 = b[tiff_fp++] & 0xff;
		if (littleEndian)
			return ((b2 << 8) + b1);
		else
			return ((b1 << 8) + b2);
	}

	
	int getValue(int fieldType, int count, byte[] b) throws IndexOutOfBoundsException{
		int value = 0;
		int unused;
		if (fieldType==SHORT && count==1) {
			value = getShort(b);
			unused = getShort(b);
		}
		else
			value = getInt(b);
		return value;
	}


	int readtiffHeader(byte[] b, int fp){
		tiff_fp = tiff_start = fp;
		int offset = -1;
		try{
			int byteOrder = getShort(b);
			if (byteOrder==0x4949) // "II"
				littleEndian = true;
			else if (byteOrder==0x4d4d) // "MM"
				littleEndian = false;
			else {
				return -1;
			}
			int magicNumber = getShort(b); // 42
			if(magicNumber != 42) 
				return -1;
			offset = getInt(b);
		}catch(IndexOutOfBoundsException e){
			return -1;
		}
		
		// This file seems to be ok; reset all image params
		tiff_width = 0;
		tiff_height = 0;
		tiff_rps = 0;
		tiff_so = null;
		tiff_sbc = null;
		return offset;
	}
	

	boolean readtiffdirectory(byte[] b, int idx) {
		tiff_fp = idx;
		int tag=0, fieldType=0, count=0, value=0;
		int nEntries=0;
		try{
			nEntries = getShort(b);
		}catch(IndexOutOfBoundsException e){
			return false;
		}
		if (nEntries<1)
			return false;
		// System.out.println("nEntries = " + nEntries);
		for (int i=0; i<nEntries; i++) {
			try{
				tag = getShort(b);
				fieldType = getShort(b);
				count = getInt(b);
				value = getValue(fieldType, count,b);
			}catch(IndexOutOfBoundsException e){
				return false;
			}
			switch (tag) {
				case IMAGE_WIDTH: 
					// System.out.println("width = " + value);
					tiff_width = value;
					break;
				case IMAGE_LENGTH: 
					// System.out.println("height = " + value);
					tiff_height = value;
					break;
				case ROWS_PER_STRIP:
					// System.out.println("rows per strip = " + value);
					tiff_rps = value;
					break;
				case STRIP_BYTE_COUNT:
					tiff_sbc = new int[count];
					if (count==1){
						// System.out.println("count = 1; stripbytecount = " + value);
						tiff_sbc[0] = value;
					}else{
						int saveLoc = tiff_fp,k;
						tiff_fp = tiff_start + value;
						for(k = 0; k < count; k++){
							try{
								tiff_sbc[k] = getInt(b);
							}catch(IndexOutOfBoundsException e){
								return false;
							}
						}
						// System.out.println("SBC: count > 1");
						tiff_fp = saveLoc;
					}
					break;
				case STRIP_OFFSETS: 
					tiff_so = new int[count];
					if (count==1){
						// System.out.println("count = 1; stripoffset = " + value);
						tiff_so[0] = value;
					}else{
						int saveLoc = tiff_fp,k;
						tiff_fp = tiff_start + value;
						for(k = 0; k < count; k++){
							try{
								tiff_so[k] = getInt(b);
							}catch(IndexOutOfBoundsException e){
								return false;
							}
						}
						// System.out.println("SO: count > 1");
						tiff_fp = saveLoc;
					}
					break;
				case BITS_PER_SAMPLE:
					// System.out.println("Pixel size = " + value);
					if(value != 8)
						return false;
					break;
				case SAMPLES_PER_PIXEL:
					/*if (value==3)
						System.out.println("RGB");
					else if (value!=1)
						System.out.println("Unsupported SamplesPerPixel: " + value);
					else
						System.out.println("1 sample per pixel");*/
					break;
				case COMPRESSION: 
					// System.out.println("Compression: " + value);
					if(value != 32773)
						return false;
					break;
				case COLOR_MAP: 
					/*if (count==768)
						System.out.println("COLOR_MAP");
					else
						System.out.println("other COLOR_MAP");*/
					break;
				default:
			}
		}
		// Sanity checks
		if( tiff_width > 0 && tiff_height > 0 && tiff_rps > 0 &&
			tiff_so != null && tiff_sbc != null && tiff_so.length == tiff_sbc.length)
			return true;
		else
			return false;
	}
	
	byte[] readtiffdata( byte[] b ){
		int i;
		byte[] u = new byte[tiff_width * tiff_height];
		int striplength = tiff_rps * tiff_width;
		
		for(i=0; i<tiff_so.length; i++){
			int i_u = i * striplength;
			if( i_u >= u.length ) return null;
			int f_u = i_u + striplength;
			if( f_u > u.length ) f_u = u.length;
			
			int i_c = tiff_start + tiff_so[i];
			if( i_c >= b.length ) return null;
			int f_c = i_c + tiff_sbc[i];
			if( f_c > b.length ) return null;

			if( !unpackstripe( b, i_c, f_c, u, i_u, f_u) ){
				System.out.println("Error unpacking strip " + i );
				return null;
			}
		}
		tiff_so = null;
		tiff_sbc = null;
		return u;
	}
	
	// return true if success
	// c - the compressed stripe
	// u - uncompressed data
	// handle data from index i to (excluding) f
	boolean unpackstripe(byte[] c, int i_c, int f_c,
						 byte[] u, int i_u, int f_u){
		byte n;
		int k;
		
		while( i_u < f_u ){
			if( i_c >= f_c ){
				// System.out.println("1");
				return false;
			}
			n = c[i_c++];
			if(n >= 0){
				k = i_c + (int)n + 1;
				if( k > f_c || i_u + (int)n + 1 > f_u ) {
					//System.out.println("2");
					return false;
				}
				while( i_c < k ){
					u[i_u++] = c[i_c++];
				}
			}else if(n==(byte)-128){
				// no operation
			}else{
				k = i_u - (int)n + 1;
				if( k > f_u ){
					//System.out.println("3");
					return false;
				}
				if( i_c >= f_c ) {
					//System.out.println("4");
					return false;
				}
				n = c[i_c++];
				while( i_u < k ){
					u[i_u++] = n ;
				}
			}
		}
		return true;
	}

	void set_qtvr_properties( Properties qtvr_properties ){
		String s = null;
		
		if(qtvr_properties == null || pv == null) return;
		
		double qtvr_width = 480.0;
		double qtvr_height = 336.0;
		
		pv.yaw = 180.0; // QTVR zero
		
		s = qtvr_properties.getProperty("width");
					if( s!=null ) { qtvr_width = Double.valueOf(s).doubleValue();}
		s = qtvr_properties.getProperty("height");
					if( s!=null ) { qtvr_height = Double.valueOf(s).doubleValue();}
		s = qtvr_properties.getProperty("fov"); // Convert VFOV->HFOV
					if( s!=null ) { double v = Double.valueOf(s).doubleValue();
									pv.hfov = (360.0/Math.PI)*Math.atan((qtvr_width/qtvr_height)*Math.tan(v*Math.PI/360.0)); }
		s = qtvr_properties.getProperty("pan");
					if( s!=null ) { pv.yaw = 180.0 - Double.valueOf(s).doubleValue(); }
		s = qtvr_properties.getProperty("tilt");
					if( s!=null ) { pv.pitch = Double.valueOf(s).doubleValue(); }

		for(int i=0; i<=pv.hotspots.size(); i++){
			s=qtvr_properties.getProperty("hotspot" + (i + 1));
			if( s != null){
				if(i<pv.hotspots.size()){ // Hotspot exists, do not overwrite
				}else{
					String t = qtvr_properties.getProperty("target" + (i + 1));
					if(t != null ){
						if( t.equalsIgnoreCase( "myself" ) ) 
							pv.hotspots.addElement("x" + i + " u'ptviewer:newPano({file=" + s + "})'");
						else
							pv.hotspots.addElement("x" + i + " u'" + s + "' t'" + t + "'");
					}else
						pv.hotspots.addElement("x" + i + " u'" + s + "'");
					//System.out.println("Added hotspot:"+ i + " " + (String)pv.hotspots.elementAt(pv.hotspots.size()-1));
				}
			}
		}
	}

	boolean check_qtvr( byte[] b ){
		byte ID[] = { (byte)0xff, (byte)0xd8, (byte)0xff, (byte)0xe0, 
					  (byte)0x00, (byte)0x10, (byte)0x4a, (byte)0x46, 
					  (byte)0x49, (byte)0x46, (byte)0x00 };
		
		if( b[4]!=(byte)'m' || b[5]!=(byte)'o' || b[6]!=(byte)'o' || b[7]!=(byte)'v' )
			return false;
			
		int	i;
		int count = b.length - ID.length;
		
		for(i=0; i<count; i++){
			if( b[i] == ID[0] && b[i+1] == ID[1] && b[i+2] == ID[2] && b[i+3] == ID[3] 
			 && b[i+4] == ID[4] && b[i+5] == ID[5] && b[i+6] == ID[6] && b[i+7] == ID[7]
		 	 && b[i+8] == ID[8] && b[i+9] == ID[9] && b[i+10] == ID[10]){ // Start of JPEG
		 	 		return true;
			}else if( b[i]==(byte)'c' && b[i+1]==(byte)'u' && b[i+2]==(byte)'b' && b[i+3]==(byte)'e'){
				hasCube = true;
			}else if( b[i]==(byte)'h' && b[i+1]==(byte)'o' && b[i+2]==(byte)'t' && b[i+3]==(byte)'t'){
				hasHotspots = true;
			}else if( b[i]==(byte)'p' && b[i+1]==(byte)'a' && b[i+2]==(byte)'n' && b[i+3]==(byte)'o'){
				hasPano = true;
			}else if( b[i]==(byte)'j' && b[i+1]==(byte)'p' && b[i+2]==(byte)'e' && b[i+3]==(byte)'g'){
				hasJpeg = true;
			}else if( b[i]==(byte)'t' && b[i+1]==(byte)'i' && b[i+2]==(byte)'f' && b[i+3]==(byte)'f'){
				hasTiff = true;
			}
		}
		return true;
	} 
				
}
